// CW_RuleMaster.m
//--------------------------------------------------------------------

#import "CW_Interface.h"
#import "CW_RuleMaster.h"

@implementation CW_RuleMaster

//--------------------------------------------------------------------

- createEnd
{

   id obj;

TRACE5(printf("\n%8X RuleMaster - entry createEnd\n",(int) self);)


// eye will be useful to locate the object during reading a storage
// dump
   eyeCatcher[0]='R';
   eyeCatcher[1]='M';
   eyeCatcher[2]='S';
   eyeCatcher[3]='T';
   obj = [super createEnd];

TRACE5(printf("%8X RuleMaster - exit createEnd %8X\n\n",
              (int) self,(int) obj);)

   return obj;

}

//--------------------------------------------------------------------

- setRuleMaker: (CW_RuleMaker *) rM
{

TRACE5(printf("\n%8X RuleMaster - entry setRuleMaker: %8X\n",
              (int) self,(int) rM);)

   ruleMaker = rM;

TRACE5(printf("%8X RuleMaster - entry setRuleMaker %8X\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- setDump: (EH_Dump *) du
{

TRACE5(printf("\n%8X RuleMaster - entry setDump: %8X\n",
              (int) self,(int) du);)

   dump = du;
//   here is a sample for usage of dump
//   [dump dumpFromAddress: self forLength: 256];

TRACE5(printf("%8X RuleMaster - entry setDump %8X\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- setShuffler: (CW_Shuffler *) sh
{

TRACE5(printf("\n%8X RuleMaster - entry setShuffler: %8X\n",
              (int) self,(int) sh);)

   shuffler = sh;

TRACE5(printf("%8X RuleMaster - entry setShuffler %8X\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- applyRuleTo: (CW_DataWarehouse *) dW with: (id) it
{

   double  sel,err,num;

TRACE5(printf("\n%8X RuleMaster - entry getAdviceFor: %8X\n",
              (int) self,(int) it);)

// We have agreed that an agent don't give immediatly his dataWarehouse
// but only his address: this way the ruleMaster will be able to decide
// what kind of information it needs and ask the agent for.

   dataWarehouse = dW;
   interface     = (CW_Interface *) it;


// First the ruleMaster need the own agent's data and provides
// translation of the agent's inputs: this phase is called detecting.

   [self keepData];
   [self detect];

   [dataWarehouse addSelection];

//To allow, statistically, each rule to fire at least once before the
//genetic algoritm and to give the auction time to select bad rules
//let's invoke the ruleMaker only if the number of selection is almost
//equal to number of rules; be carefull: if you remove the first part
//of test you will have zero divisition after application of evolution
//because the selAfterEvol counter is set to zero.
//If you needs more precise verification you can reduce the ratio
//in the second part of test.

   sel = (double) selAfterEvol;
   err = (double) errorCounter;
   num = (double) [ruleList getCount];

   if (((evolutionRate == 1) && (sel > num)) ||
       ((evolutionRate < 1) &&
     ([aUniformDblRand getDoubleWithMin: 0 withMax: 1] < evolutionRate)))
   {

      if ((err/sel) > confidence)
      {

      [ruleMaker evolveDataWarehouse: dataWarehouse];
      [dataWarehouse addEvolution];

         //To avoid problems if a rule that have fired a message or
         //that is waiting for a reward from effector will be dropped
         //by ruleMaker let's clean messageList and give the effectors
         //order to reset theyre partnerList
         //Because we clean messageList the action of detect must be
         //repeated; also keepData must be redone because the ruleMaker
         //can modify some of the data in dataWarehouse.

         [effectorList forEach: M(resetPartnerList)];
         [self keepData];
         [self detect];

      }

   }

   [self matchRule];

//Following loop is intended to guarantee a match, if we haven't matches
//the most similar rule is asked to change the value of a non-wildkard
//allele, random chosen.

   if ([matchList getCount] == 0)
   {

      [self coverDetector];
      [self matchRule];

   }
//Having matches we can do an auction to select rules going to fire
//after that the rule have to pay some taxes.

   [self doAnAuction];
   [self collectTaxes];
   [self fire];

//The effectors try to match new messages, if some matches are done
//choose method will be going to select one of them. All the messages
//that match effectors are removed from the message list; remaining
//messages will be matched from rules in the next loop.

   [self matchEffector];

   if ([matchList getCount] == 0) [self coverEffector];

   [self choose];

//When a suggestion is made, the number of the action is returned to the
//agent and his value is stored into agent's dataWarehouse; note that is
//stored also the current bad answers counter.

   [dataWarehouse setActiveEffector: activeEffector];

TRACE5(printf("\n%8X RuleMaster - exit getAdviceAbout %d\n\n",
              (int) self,suggestion);)

   [interface setSuggestionMap: [activeEffector getBody]];
   return self;

}

//--------------------------------------------------------------------

- keepData
{

   int i = 0;

TRACE5(printf("\n%8X RuleMaster - entry keepData\n",(int) self);)

   // phase 0 - Housekeeping: asking the dataWarehouse about
   //           lists and objects

   aUniformIntRand = [dataWarehouse getUniformIntRand];
   aUniformDblRand = [dataWarehouse getUniformDblRand];

   parmWarehouse  = [dataWarehouse  getParmWarehouse];
   treasury       = [dataWarehouse  getTreasury];

   ruleList       = [dataWarehouse  getRuleList];
   effectorList   = [dataWarehouse  getEffectorList];
   messageList    = [dataWarehouse  getMessageList];
   matchList      = [dataWarehouse  getMatchList];
   winnerList     = [dataWarehouse  getWinnerList];
   workList       = [dataWarehouse  getWorkList];
   memoList       = [dataWarehouse  getMemoList];
   errorCounter   = [dataWarehouse  getErrorCounter];
   selAfterEvol   = [dataWarehouse  getSelAfterEvol];

   // then let's ask classifierParm for parameters

   if ([parmWarehouse getParm: "geneLength"
                  intoInteger: &geneLength] != 0)
   {
      printf("CW_RuleMaster cannot find geneLength\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "maxNumberOfMessages"
                  intoInteger: &maxNumberOfMessages] != 0)
   {
      printf("CW_RuleMaster cannot find maxNumberOfMessages\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "maxMemory"
                  intoInteger: &maxMemory] != 0)
   {
      printf("CW_RuleMaster cannot find maxMemory\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "shuffleFlag"
                  intoInteger: &shuffleFlag] != 0)
   {
      printf("CW_RuleMaster cannot find shuffleFlag value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "bidRatio"
                    intoFloat: &bidRatio] != 0)
   {
      printf("CW_RuleMaster cannot find bidRatio value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "linearBid1"
                    intoFloat: &lBid1] != 0)
   {
      printf("CW_RuleMaster cannot find linearBid1 value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "linearBid2"
                    intoFloat: &lBid2] != 0)
   {
      printf("CW_RuleMaster cannot find linearBid2 value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "effectiveLinearBid1"
                    intoFloat: &eLBid1] != 0)
   {
      printf("CW_RuleMaster cannot find effectiveLinearBid1 value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "effectiveLinearBid2"
                    intoFloat: &eLBid2] != 0)
   {
      printf("CW_RuleMaster cannot find effectiveLinearBid2 value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "bidSigma"
                    intoFloat: &bidSigma] != 0)
   {
      printf("CW_RuleMaster cannot find bidSigma value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "bidMu"
                    intoFloat: &bidMu] != 0)
   {
      printf("CW_RuleMaster cannot find bidMu value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "evolutionRate"
                    intoFloat: &evolutionRate] != 0)
   {
      printf("CW_RuleMaster cannot find evolutionRate value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "confidence"
                    intoFloat: &confidence] != 0)
   {
      printf("CW_RuleMaster cannot find confidence value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "lifeTaxRate"
                    intoFloat: &lifeTaxRate] != 0)
   {
      printf("CW_RuleMaster cannot find lifeTaxRate value\n");
      i = 1;
   }

   if ([parmWarehouse getParm: "bidTaxRate"
                    intoFloat: &bidTaxRate] != 0)
   {
      printf("CW_RuleMaster cannot find bidTaxRate value\n");
      i = 1;
   }

   if (i != 0)
   {
      printf("CW_RuleMaster cannot find one or more parameters\n");
      exit (0);
   }

TRACE5(printf("%8X RuleMaster - exit keepData %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------
//Here we encode agent's information: for each event we can know that
//it is true or false or, may be, we don't know anything about it (#).
//encoding the information a new message is generated and put into the
//message list.
- detect
{

   char       * string;
   CW_Message * message;
   CW_Message * memo;
   int          i,j=0,k=0;

TRACE5(printf("\n%8X RuleMaster - entry detect\n",(int) self);)

   string = [interface getEventsMap];

   message = [CW_Message createBegin: [self getZone]];
             [message    setText:     &string[0] length: geneLength
                         owner:       treasury   dump:   dump];
   message = [message    createEnd];
   [messageList addLast: message];

   if (maxMemory > 0)
     {
       j = [memoList getCount];
       for (i=0;i<j;i++)
	 {
	   memo = [memoList atOffset: i];
	   if ([memo compareWith: &string[0]] == 0) 
	     {
	       [dataWarehouse setLastMemo: memo];
	       k = 1;
	     }
	 }

       if (k == 0)
	 {
	   memo    = [CW_Message createBegin: [self getZone]];
	   [memo       setText:     &string[0] length: geneLength
		       owner:       treasury   dump:   dump];
	   memo    = [memo       createEnd];
	   [memoList addLast: memo];
	   [dataWarehouse setLastMemo: memo];
	   if ([memoList getCount] > maxMemory)
	     {
	       memo = [memoList removeFirst];
	       [memo drop];
	     }
	 }
     }

TRACE5(printf("%8X RuleMaster - exit detect %8X\n\n",
        (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- matchRule
{

   int        i,j,numberOfRules,numberOfMessages,match,totalMatches;
   int        firstTime=0;
   double     strength,worstStrength;
   CW_Message * message;
   CW_Rule    * rule;

TRACE5(printf("\n%8X RuleMaster - entry matchRule\n",(int) self);)

   //to store new matches a new match list is created after destroing
   //the old one

   [matchList removeAll];

   if (shuffleFlag == 1)
   {
      [shuffler shuffle: messageList using: aUniformIntRand];
      [shuffler shuffle: ruleList    using: aUniformIntRand];
   }

   numberOfRules    = [ruleList    getCount];
   numberOfMessages = [messageList getCount];
   worstStrength    = 100;

   for (i=0;i<numberOfRules;i++)
   {

      rule = [ruleList atOffset: i];
      [rule resetPartnerList];
      totalMatches = 0;

      for (j=0;j<numberOfMessages;j++)
      {

         message     = [messageList atOffset: j];

         match = [rule matchMessage: message];
         strength = [rule getStrength];

         if (match == geneLength) totalMatches++;

         if ((strength < worstStrength) || (firstTime == 0))
         {

            worstStrength = strength;
            ruleWithBestMatch = rule;

         }

         if (firstTime == 0) firstTime = 1;

      }

      if (totalMatches > 0) [matchList addLast: rule];

   }

TRACE5(printf("%8X RuleMaster - exit matchRule %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- coverDetector
{

   char * string;
   int i;

TRACE5(printf("\n%8X RuleMaster - entry coverDetector\n",(int) self);)

   [dataWarehouse addCoverDetector];
   i = [aUniformIntRand getIntegerWithMin: 0 withMax:
        [messageList getCount]-1];

   string = [[messageList atOffset: i] getText];

   [ruleWithBestMatch modifyConditionTo: string];


TRACE5(printf("%8X RuleMaster - exit coverDetector %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- coverEffector
{

   char        * string;
   CW_Effector * anEffector;
   int         i;

TRACE5(printf("\n%8X RuleMaster - entry coverEffector\n",(int) self);)

   [dataWarehouse addCoverEffector];
   i = [aUniformIntRand getIntegerWithMin: 0 withMax:
        [effectorList getCount]-1];

   anEffector = [effectorList atOffset: i];

   string = [anEffector getBody];

   [ruleWithBestAction modifyActionTo: string];
   [matchList addLast: anEffector];
   [anEffector setPartner: ruleWithBestAction];
   [messageList remove: bestMessage];
   [bestMessage drop];

TRACE5(printf("%8X RuleMaster - exit coverEffector %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- matchEffector
{

   int         i,j,numberOfEffectors,numberOfMessages,match;
   int         totalMatches,numberOfMessageToRemove,firstTime=0;
   double      strength,worstStrength;
   CW_Effector * effector;
   CW_Message  * message;
   CW_Rule     * rule;

TRACE5(printf("\n%8X RuleMaster - entry matchEffector\n",(int) self);)

   //to store new matches a new match list is created after destroing
   //the old one

   [matchList removeAll];

   if (shuffleFlag == 1)
   {
      [shuffler shuffle: messageList  using: aUniformIntRand];
      [shuffler shuffle: effectorList using: aUniformIntRand];
   }

   bestMatch = -1;
   worstStrength = 100;
   numberOfEffectors = [effectorList getCount];
   for (i=0;i<numberOfEffectors;i++)
   {

      effector = [effectorList atOffset: i];
      [effector resetPartnerList];
      totalMatches = 0;

      [workList removeAll];

      numberOfMessages = [messageList  getCount];
      for (j=0;j<numberOfMessages;j++)
      {

         message     = [messageList atOffset: j];
         rule        = [message getOwner];
         strength    = [rule getStrength];

         match = [effector matchMessage: message];

         if ((strength < worstStrength) || (firstTime == 0))
         {

            bestMessage        = message;
            ruleWithBestAction = rule;
            worstStrength      = strength;

         }

         if (firstTime == 0) firstTime = 1;

         if (match >= geneLength)
         {

            totalMatches++;
            [workList addLast: message];

         }

      }

      if (totalMatches > 0) [matchList addLast: effector];

      numberOfMessageToRemove = [workList getCount];
      for(j=0;j<numberOfMessageToRemove;j++)
      {

         message = [workList atOffset: j];
         [messageList remove: message];
         [message drop];

      }

      [workList removeAll];

   }

TRACE5(printf("%8X RuleMaster - exit matchEffector %8X\n\n",
       (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- doAnAuction
{

   CW_Rule * rule;
   CW_Rule * winner=nil;
   int     i,j,numberOfMatches,maxMsg;
   double  maxBid,ruleBid,sel,err;

TRACE5(printf("\n%8X RuleMaster - entry doAnAuction\n",(int) self);)

   numberOfMatches = [matchList getCount];

//First matching rules are asked to bid basing on the bid's parameters
//contained into the classifierParm.

   maxMsg = maxNumberOfMessages;
   sel = (double) selAfterEvol + 1;
   err = (double) errorCounter + 1;
   if ((err/sel) < confidence) bidSigma = 0;

   for(i=0;i<numberOfMatches;i++)
   {

      rule = [matchList atOffset: i];
      [rule bidWithRatio: bidRatio lBid1: lBid1 lBid2: lBid2
              eLBid1: eLBid1  eLBid2: eLBid2 bidSigma: bidSigma
               bidMu: bidMu];

   }

   [winnerList removeAll];

//then several rules are chosen to fire accordingly with theyre
//bid value.

   do
   {

      maxBid = -1;
      for(j=0;j<numberOfMatches;j++)
      {

         rule = [matchList atOffset: j];
         ruleBid = [rule getEBid];
         if (ruleBid > maxBid)
         {

            maxBid = ruleBid;
            winner = rule;

         }

      }

      [winnerList addLast: winner];
      [matchList  remove:  winner];
      numberOfMatches--;
      maxMsg--;

   } while ((maxMsg > 0) && (numberOfMatches > 0));

TRACE5(printf("%8X RuleMaster - exit doAnAuction %8X\n\n",
              (int) self,(int) self);)

   return self;
}

//--------------------------------------------------------------------

- collectTaxes
{

   int     i,numberOfRules;
   CW_Rule * rule;

TRACE5(printf("\n%8X RuleMaster - entry collectTaxes\n",(int) self);)

   // first let's collect lifeTax

   numberOfRules = [ruleList getCount];
   for (i=0;i<numberOfRules;i++)
   {

      rule = [ruleList atOffset: i];
      [rule payTaxOf: lifeTaxRate to: treasury];
//      if ([rule getStrength] > 2) [rule setStrength: 2];

   }

   // then matching rules have to pay bid tax, remember that the list
   // has been divided between match and winner

   numberOfRules = [matchList getCount];
   for (i=0;i<numberOfRules;i++)
   {

      rule = [matchList atOffset: i];
      [rule payTaxOf: bidTaxRate to: treasury];

   }

   numberOfRules = [winnerList getCount];
   for (i=0;i<numberOfRules;i++)
   {

      rule = [winnerList atOffset: i];
      [rule payTaxOf: bidTaxRate to: treasury];

   }

TRACE5(printf("%8X RuleMaster - exit collectTaxes %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- fire
{

   int        s,i,numberOfWinners;
   char       * text;
   CW_Message * message;
   CW_Rule * rule;

TRACE5(printf("\n%8X RuleMaster - entry fire\n",(int) self);)

   s = [dataWarehouse getSelectionCounter];

   [messageList forEach: M(drop)];
   [messageList removeAll];

   numberOfWinners = [winnerList getCount];

   for (i=0;i<numberOfWinners;i++)
   {

      rule = [winnerList atOffset: i];
      text = [rule payAndFire];

      message = [CW_Message createBegin: [self getZone]];
                [message    setText:     text length: geneLength
                            owner:       rule dump:   dump];
      message = [message    createEnd];

      [messageList addLast: message];

   }

TRACE5(printf("%8X RuleMaster - exit fire %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- choose
{

   int         i,numberOfMatches;
   double      maxSupport=0,effectorSupport;
   CW_Effector * effector;

TRACE5(printf("\n%8X RuleMaster - entry choose\n",(int) self);)

   suggestion      = -1;
   numberOfMatches = [matchList getCount];

   for (i=0;i<numberOfMatches;i++)
   {

      effector = [matchList atOffset: i];
      effectorSupport = [effector getSupport];
      if (effectorSupport > maxSupport)
      {

         maxSupport = effectorSupport;
         activeEffector = effector;

      }

      suggestion = [effector getSuggestion];

   }

TRACE5(printf("%8X RuleMaster - exit choose %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------

- setReward: (double) rV to: (CW_DataWarehouse *) dW
{

   CW_Effector * effector;
   CW_Message  * memo; 
   float effectiveReward=0;

TRACE5(printf("\n%8X RuleMaster - entry setReward: %f\n",(int) self,rV);)
TRACE5(printf("                                  for: %8X\n",(int) dW);)

   dataWarehouse = dW;

   parmWarehouse  = [dataWarehouse getParmWarehouse];
   effector       = [dataWarehouse getActiveEffector];
   treasury       = [dataWarehouse getTreasury];

   if ([parmWarehouse getParm: "maxMemory"
                  intoInteger: &maxMemory] != 0)
   {
      printf("CW_RuleMaster cannot find maxMemory\n");
      exit(0);
   }

   effectiveReward = rV;
   if (maxMemory > 0)
     {
       memo          = [dataWarehouse getLastMemo];
       effectiveReward = [memo setSupport: rV];
     }

     if(effectiveReward == 0) [dataWarehouse addError];
     [treasury pay: effectiveReward];
     [effector payToPartners: (double) effectiveReward];

TRACE5(printf("%8X RuleMaster - exit setReward: %8X\n\n",
              (int) self,(int) self);)

   return self;

}

//--------------------------------------------------------------------
@end
